#pragma once

#include <libc/component.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <netdb.h>
#include <util/list.h>
#include <base/heap.h>
#include "command.h"
#include "error.h"
#include "exception.h"
#include "payload.h"
#include "message.h"
#include "sha256.h"

namespace Hsm {
  class Connection;

	using Hostname = Genode::String<256>;
	using Ip_address = Genode::String<64>;
	using Username = Genode::String<256>;
	using Password = Genode::String<256>;
	using Key_name = Genode::String<256>;
	using Crypto_method_label = Genode::String<256>;

	Crypto_method method_from_label(const Crypto_method_label label)
	{
		if (label == Crypto_method_label("EcDsaNone")) {
			return Crypto_method::EcDsaNone;
		} else {
			Genode::error("Unknown crypto method ", label);
			throw Invalid_argument_exception(); 
		}
	}
}

class Hsm::Connection
{
	private:
		enum {
			INVALID_SOCKET = -1,
			SALT_SIZE = 32,
			HASH_SIZE = 32,
			MAGIC_SIZE = 32,
		};

		Genode::Heap& _heap;
		int _socket = INVALID_SOCKET;

		Ip_address resolve_hostname(const Hostname& hostname) const
		{
			struct hostent *host_entry;
			struct in_addr **addr_list;
			int i;

			if ((host_entry = gethostbyname(hostname.string())) == NULL) {
				Genode::error("gethostbyname for ", hostname, " failed");
				throw Genode::Exception();
			}

			addr_list = (struct in_addr **)host_entry->h_addr_list;

			for(i = 0; addr_list[i] != NULL; i++) {
				return Ip_address(static_cast<const char*>(inet_ntoa(*addr_list[i])));
			}

			Genode::error("no ip address for ", hostname, " found");
			throw Genode::Exception();
		}

		void _login_request(const Username& username,
		                    char* client_login_salt,
		                    char* server_login_salt) const
		{
			Message request(_heap, Command::CommandLoginRequest);
			request.add(Data_type::typeIdLabel, username.string());
			_get_random(client_login_salt, SALT_SIZE);
			request.add(Data_type::typeRandomData, client_login_salt, SALT_SIZE);
			request.add(Data_type::typeUserId, username.string());

			Message& response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			auto server_salt_payload = response.payload_by_type(Data_type::typeLoginData);
			if (server_salt_payload == nullptr) {
				Genode::error("Server salt missing");
				throw Missing_payload_exception();
			}
			Genode::memcpy(server_login_salt, server_salt_payload->payload_buffer(), SALT_SIZE);
			destroy(_heap, &response);
		}

		void _compute_challenge(const void* magic,
		                        const void* salt1,
		                        const void* salt2,
		                        const char* password,
		                        char* challenge) const
		{
			char hash_buffer[MAGIC_SIZE + (2 * SALT_SIZE) + HASH_SIZE];
			Genode::memcpy(hash_buffer, magic, MAGIC_SIZE);
			Genode::memcpy(hash_buffer + MAGIC_SIZE, salt1, SALT_SIZE);
			Genode::memcpy(hash_buffer + MAGIC_SIZE + SALT_SIZE, salt2, SALT_SIZE);
			sha256(password, strlen(password), hash_buffer + MAGIC_SIZE + (2 * SALT_SIZE), HASH_SIZE);
			sha256(hash_buffer, sizeof(hash_buffer), challenge, HASH_SIZE);
		}

		bool _login_response(const Password& password,
		                     const char* client_login_salt,
		                     const char* server_login_salt) const
		{
			char request_hash[HASH_SIZE];
			_compute_challenge(magic_login_request, server_login_salt,
			                   client_login_salt, password.string(), request_hash);

			Message request(_heap, Command::CommandLoginResponse);
			request.add(Data_type::typeLoginData, request_hash, sizeof(request_hash));
			char client_auth_salt[SALT_SIZE];
			_get_random(client_auth_salt, sizeof(client_auth_salt));
			request.add(Data_type::typeRandomData, client_auth_salt, sizeof(client_auth_salt));

			Message& response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			auto server_salt_payload = response.payload_by_type(Data_type::typeRandomData);
			if (server_salt_payload == nullptr) {
				Genode::error("Server salt missing");
				throw Missing_payload_exception();
			}
			auto client_challenge = response.payload_by_type(Data_type::typeLoginData);
			if (client_challenge == nullptr) {
				Genode::error("Client_challenge missing");
				throw Missing_payload_exception();
			}
			char auth_hash[HASH_SIZE];
			_compute_challenge(magic_login_auth, server_salt_payload->payload_buffer(),
			                   client_auth_salt, password.string(), auth_hash);
			bool login_successful = (Genode::memcmp(auth_hash, client_challenge->payload_buffer(), SALT_SIZE) == 0);

			destroy(_heap, &response);
			return login_successful;
		}

		void _get_random(void* buffer, const uint32_t length) const
		{
			Genode::memset(buffer, 0x23, length);
			Genode::warning(__func__, " not implemented");
		}

	public:
		Connection(Genode::Heap& heap, const Ip_address& ip_address, const uint16_t port)
			: _heap(heap)
		{
			Libc::with_libc([&] () {
				Genode::log("connecting to ", ip_address, ":", port);

				_socket = socket(AF_INET , SOCK_STREAM , 0);
				if (_socket == INVALID_SOCKET) {
					Genode::error("socket failed");
					throw Genode::Exception();
				}

				struct sockaddr_in address;
				address.sin_addr.s_addr = inet_addr(ip_address.string());
				address.sin_family = AF_INET;
				address.sin_port = htons(port);

				if (connect(_socket, (struct sockaddr *)&address, sizeof(address)) < 0) {
					Genode::error("connect failed");
					throw Genode::Exception();
				}

				Genode::log("connected");
			});
		}

		Message& execute(const Message& request) const
		{
			request.write_to_socket(_socket);
			auto response = new (_heap) Message(_heap, _socket);

			if (response->status() >= SECUROSYS_ERROR_BASE) {
				Genode::error("HSM replied with error ", Genode::Hex(response->status()));
				throw Hsm::Error_response_exception();
			}

			return *response;
		}

		void hello() const
		{
			Message request(_heap, Command::CommandHello);
			request.add(Data_type::typeMagic, LOGIN_MAGIC);

			Message& response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			destroy(_heap, &response);
		}

		bool login(const Username& username, const Password& password) const
		{
			char client_salt[32];
			char server_salt[32];
			_login_request(username, client_salt, server_salt);
			return _login_response(password, client_salt, server_salt);
		}

		uint32_t sign(const Crypto_method method,
		              const Key_name& key_name,
		              const void* message_data,
		              const uint32_t message_size,
		              void* signature_data,
		              const uint32_t signature_size) const
		{
			Message request(_heap, Command::CommandAsymmetricSingleSign);
			request.add(Data_type::typeCryptoMethod, static_cast<uint32_t>(method));
			request.add(Data_type::typeIdLabel, key_name.string());
			request.add(Data_type::typeSignData, message_data, message_size);

			Message& response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			auto signature = response.payload_by_type(Data_type::typeSignatureData);
			if (signature == nullptr) {
				Genode::error("Signature data missing");
				throw Missing_payload_exception();
			}
			if (signature_size < signature->payload_length()) {
				Genode::error("Signature buffer to small");
				throw Buffer_too_short_exception();
			}
			const uint32_t output_size = signature->payload_length();
			Genode::memcpy(signature_data, signature->payload_buffer(), output_size);

			destroy(_heap, &response);

			return output_size;
		}

		bool verify(const Crypto_method method,
		            const Key_name& key_name,
		            const void* message_data,
		            const uint32_t message_size,
		            const void* signature_data,
		            const uint32_t signature_size) const
		{
			Message request(_heap, Command::CommandAsymmetricSingleVerify);
			request.add(Data_type::typeCryptoMethod, static_cast<uint32_t>(method));
			request.add(Data_type::typeIdLabel, key_name.string());
			request.add(Data_type::typeVerifyData, message_data, message_size);
			request.add(Data_type::typeSignatureData, signature_data, signature_size);

			Message& response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			bool signature_good = false;

			switch (response.status())
			{
				case CKR_OK:
					signature_good = true;
					break;
				case CKR_SIGNATURE_INVALID:
					signature_good = false;
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			destroy(_heap, &response);

			return signature_good;
		}
};

